# - CMAKE Config file for the MADNESS package
#
# Valid COMPONENTS: world tensor misc tensor linalg tinyxml muparser mra chem
#
# The following variables are defined:
#  madness_FOUND                - System has the MADNESS package
#  madness_INCLUDE_DIRS         - The MADNESS include directory
#  madness_LIBRARIES            - The MADNESS libraries and their dependencies
#  madness_INSTALL_LIBRARY_DIR  - Where MADNESS libraries will be installed (same for CONFIG files imported from the build and install trees)
#  madness_<COMPONENT>_FOUND    - System has the specified MADNESS COMPONENT
#  madness_<COMPONENT>_LIBRARY  - The MADNESS COMPONENT library
#  madness_F77_INTEGER_SIZE     - The default F77 integer size used for BLAS calls
#  madness_ASSUMES_ASLR_DISABLED - Whether MADNESS assumes ASLR to be disabled
#
# Additionally MADNESS exports the following variables from its dependent packages:
# = LAPACK =
#    LAPACK_LIBRARIES           Libraries providing LAPACK and its dependents
#    LAPACK_INCLUDE_DIRS        Include directories where to locate the header files needed to use LAPACK_LIBRARIES
#    LAPACK_COMPILE_DEFINITIONS Compile definitions needed to use the LAPACK header files
#    LAPACK_COMPILE_OPTIONS     Compile options needed to use the LAPACK header files
# = MPI =
#    MPIEXEC_EXECUTABLE         Executable for running MPI programs
#    MPIEXEC_NUMPROC_FLAG       Flag to pass to MPIEXEC_EXECUTABLE before giving
#                               it the number of processors to run on
#    MPIEXEC_PREFLAGS           Flags to pass to MPIEXEC_EXECUTABLE directly
#                               before the executable to run.
#    MPIEXEC_POSTFLAGS          Flags to pass to MPIEXEC_EXECUTABLE after other flags

# Set package version
set(madness_VERSION "0.10.1")


####### Expanded from @PACKAGE_INIT@ by configure_package_config_file() #######
####### Any changes to this file will be overwritten by the next CMake run ####
####### The input file was madness-config.cmake.in                            ########

get_filename_component(PACKAGE_PREFIX_DIR "${CMAKE_CURRENT_LIST_DIR}/../../../" ABSOLUTE)

macro(set_and_check _var _file)
  set(${_var} "${_file}")
  if(NOT EXISTS "${_file}")
    message(FATAL_ERROR "File or directory ${_file} referenced by variable ${_var} does not exist !")
  endif()
endmacro()

macro(check_required_components _NAME)
  foreach(comp ${${_NAME}_FIND_COMPONENTS})
    if(NOT ${_NAME}_${comp}_FOUND)
      if(${_NAME}_FIND_REQUIRED_${comp})
        set(${_NAME}_FOUND FALSE)
      endif()
    endif()
  endforeach()
endmacro()

####################################################################################

set(madness_CMAKE_TOOLCHAIN_FILE "")

# ensure that CMAKE_SYSTEM_NAME matches what was used to compile MADNESS
if (NOT CMAKE_SYSTEM_NAME STREQUAL Darwin)
  set (ERROR_MESSAGE "CMAKE_SYSTEM_NAME=${CMAKE_SYSTEM_NAME}, does not match CMAKE_SYSTEM_NAME=Darwin used to compile MADNESS")
  if (madness_CMAKE_TOOLCHAIN_FILE)
    set (ERROR_MESSAGE "${ERROR_MESSAGE}; make sure to use toolchain  to compile your code!")
    message (FATAL_ERROR "${ERROR_MESSAGE}")
  endif(madness_CMAKE_TOOLCHAIN_FILE)
endif()

# check for prereqs:
# 1. Threads
if (NOT TARGET Threads::Threads)
  set(CMAKE_THREAD_PREFER_PTHREAD TRUE)
  find_package(Threads REQUIRED)
endif()

# 2. Cereal
set(madness_HAS_CEREAL )
if (madness_HAS_CEREAL)
  get_filename_component(cereal_DIR "" DIRECTORY)
  find_package(cereal QUIET CONFIG HINTS "${cereal_DIR}")
  if (NOT TARGET cereal)
    add_library(cereal INTERFACE IMPORTED)
    set_target_properties(cereal PROPERTIES
          INTERFACE_INCLUDE_DIRECTORIES "${cereal_INCLUDE_DIRS}"
          INTERFACE_COMPILE_DEFINITIONS "CEREAL_THREAD_SAFE=1")
  endif(NOT TARGET cereal)
endif(madness_HAS_CEREAL)

# Set the madness libraries target
set(madness_LIBRARIES madness)

set(madness_SOURCE_DIR "/opt/local/var/macports/build/_opt_local_ppcports_science_madness/madness/work/madness-eb72069dd37f6a302561b5d9743931fdb2fb2dc8")
set(madness_BINARY_DIR "/opt/local/var/macports/build/_opt_local_ppcports_science_madness/madness/work/build")

# Set Elemental variables
set(madness_HAS_ELEMENTAL 0)
set(madness_HAS_ELEMENTAL_EMBEDDED )

set(madness_INSTALL_LIBRARY_DIR "${PACKAGE_PREFIX_DIR}/lib")
set(madness_BUILD_INCLUDE_DIRS "${madness_SOURCE_DIR}/src" ":" "${madness_BINARY_DIR}/src")
set(madness_INSTALL_INCLUDE_DIRS "${PACKAGE_PREFIX_DIR}/include" ":"
      "${PACKAGE_PREFIX_DIR}/include/madness/external")
set(madness_INSTALL_CMAKEDIR "${PACKAGE_PREFIX_DIR}/lib/cmake/madness")

if (madness_HAS_ELEMENTAL)
  if(madness_HAS_ELEMENTAL_EMBEDDED)
    set(Elemental_SOURCE_DIR "")
    set(Elemental_BINARY_DIR "")
    list(APPEND madness_BUILD_INCLUDE_DIRS
        "${Elemental_SOURCE_DIR}/include" "${Elemental_BINARY_DIR}/include"
        "${Elemental_SOURCE_DIR}/external/pmrrr/include"
        "${Elemental_BINARY_DIR}/external/pmrrr/include")
  else(madness_HAS_ELEMENTAL_EMBEDDED)
    set(Elemental_SOURCE_DIR "")
    set(Elemental_BINARY_DIR "")
    # elemental package and component name depends on version
    set(ELEMENTAL_PACKAGE_NAME )
    set(ELEMENTAL_CONFIGURE_NAME )
    # where to look for elemental cmake config file
    if(CMAKE_CURRENT_LIST_DIR EQUAL madness_BINARY_DIR)
      set(Elemental_DIR "${Elemental_BINARY_DIR}")
    else()
      set(Elemental_DIR "${PACKAGE_PREFIX_DIR}/CMake/elemental")
    endif()
    find_package(${ELEMENTAL_CONFIGURE_NAME} REQUIRED
                 COMPONENTS REQUIRED ${ELEMENTAL_PACKAGE_NAME})
    list(APPEND madness_BUILD_INCLUDE_DIRS
        "${Elemental_SOURCE_DIR}/include" "${Elemental_BINARY_DIR}/include"
        "${Elemental_SOURCE_DIR}/external/pmrrr/include"
        "${Elemental_BINARY_DIR}/external/pmrrr/include"
        "${Elemental_SOURCE_DIR}/external/pmrrr/suite_sparse"
        "${Elemental_BINARY_DIR}/external/pmrrr/suite_sparse"
        )
  endif(madness_HAS_ELEMENTAL_EMBEDDED)
endif(madness_HAS_ELEMENTAL)

# define madness_INCLUDE_DIRS according to where we are compiling: MADNESS build tree or outside
# external packages should use madness_BUILD_INCLUDE_DIRS and madness_INSTALL_INCLUDE_DIRS directly
if(CMAKE_CURRENT_LIST_DIR EQUAL madness_BINARY_DIR)
  set(madness_INCLUDE_DIRS "${madness_BUILD_INCLUDE_DIRS}")
else()
  set(madness_INCLUDE_DIRS "${madness_INSTALL_INCLUDE_DIRS}")
endif()

# Set the Fortran 77 integer size (4 or 8)
set(madness_F77_INTEGER_SIZE 4)

# whether MADNESS assumes ASLR to be disabled
set(madness_ASSUMES_ASLR_DISABLED "OFF")

# list of dependencies of MADworld
set(madness_MADWORLD_INTERFACE_LINK_LIBRARIES "Threads::Threads")

########### LAPACK ############
# this is formatted to be added to CMAKE_REQUIRED_LIBRARIES
set(LAPACK_LIBRARIES "OpenBLAS")
set(LAPACK_INCLUDE_DIRS "")
set(LAPACK_COMPILE_DEFINITIONS "")
set(LAPACK_COMPILE_OPTIONS "")

########### MPI ############
set(MPIEXEC_EXECUTABLE "" CACHE FILEPATH "Executable for running MPI programs")
set(MPIEXEC_NUMPROC_FLAG "" CACHE STRING "Maximum number of processors available to run MPI applications")
set(MPIEXEC_PREFLAGS "" CACHE STRING "Flag used by MPI to specify the number of processes for mpiexec")
set(MPIEXEC_POSTFLAGS "" CACHE STRING "These flags will be placed after all flags passed to mpiexec")
# load MPI::MPI_CXX ... also, if MADworld depends on PaRSEC, enable C language and load MPI_C
cmake_policy(SET CMP0057 NEW)  # treat IN_LIST as operator
if (MPI::MPI_CXX IN_LIST madness_MADWORLD_INTERFACE_LINK_LIBRARIES)
    set(_mpi_languages CXX)
endif()
if (PaRSEC::parsec IN_LIST madness_MADWORLD_INTERFACE_LINK_LIBRARIES)
   enable_language(C)
   list(APPEND _mpi_languages C)
endif()
find_package(MPI REQUIRED COMPONENTS "${_mpi_languages}")

########### PaRSEC ############
if (PaRSEC::parsec IN_LIST madness_MADWORLD_INTERFACE_LINK_LIBRARIES AND NOT TARGET PaRSEC::parsec)
    include( CMakeFindDependencyMacro )
    get_filename_component(_parsec_config_dir "" DIRECTORY)
    find_dependency(PaRSEC CONFIG REQUIRED COMPONENTS parsec PATHS "${_parsec_config_dir}" NO_DEFAULT_PATH)
endif()

########### TBB ############
if (TBB::tbb IN_LIST madness_MADWORLD_INTERFACE_LINK_LIBRARIES)
   if (NOT TARGET TBB::tbb)
       include( CMakeFindDependencyMacro )
       if (NOT "" STREQUAL "")
           get_filename_component(_tbb_config_dir "" DIRECTORY)
           find_dependency(TBB CONFIG REQUIRED COMPONENTS tbb PATHS "${_tbb_config_dir}" NO_DEFAULT_PATH)
       else()
           # use madness' FindTBB
           set(_orig_CMAKE_MODULE_PATH "${CMAKE_MODULE_PATH}")
           if(CMAKE_CURRENT_LIST_DIR EQUAL madness_BINARY_DIR)
               set(madness_MODULE_DIR "${madness_SOURCE_DIR}/cmake/modules")
           else()
               set(madness_MODULE_DIR "${madness_INSTALL_CMAKEDIR}/modules")
           endif()
           set(CMAKE_MODULE_PATH "${madness_MODULE_DIR};${CMAKE_MODULE_PATH}")
           get_filename_component(_tbb_root_dir "/.." DIRECTORY)
           find_dependency(TBB REQUIRED COMPONENTS tbb PATHS "${_tbb_root_dir}" NO_DEFAULT_PATH)
           set(CMAKE_MODULE_PATH "${_orig_CMAKE_MODULE_PATH}")
       endif()
   endif()
endif()

########### LibXC ############
if ()
  #message("LIBXC_ROOT_DIR" "${LIBXC_ROOT_DIR}")
  set(LIBXC_ROOT_DIR "/..")
  #message("LIBXC_ROOT_DIR" "${LIBXC_ROOT_DIR}")
  set(LIBXC_INCLUDE_DIRS "${LIBXC_INCLUDE_DIRS}" ":" "")
  set(LIBXC_LIBRARIES "")
  set(LIBXC_FOUND 1)
endif()
  
### LAPACK_LIBRARIES might include IMPORTED targets that are not globally available
if (LAPACK_LIBRARIES MATCHES OpenMP::OpenMP_C AND NOT TARGET OpenMP::OpenMP_C)
  find_package(OpenMP REQUIRED COMPONENTS C)
endif()
if (LAPACK_LIBRARIES MATCHES Threads::Threads AND NOT TARGET Threads::Threads)
  find_package(Threads REQUIRED)
endif()

# Include library IMPORT targets
if(NOT TARGET madness)
  include("${CMAKE_CURRENT_LIST_DIR}/madness-targets.cmake")
endif(NOT TARGET madness)

# Set component variables
foreach(_comp ${madness_FIND_COMPONENTS})
  if(("${_comp}" STREQUAL madness AND TARGET madness) OR TARGET MAD${_comp})
    set(madness_${_comp}_FOUND TRUE)
    set(madness_${_comp}_LIBRARY MAD${_comp})
  else()
    set(madness_${_comp}_FOUND FALSE)
    set(madness_${_comp}_LIBRARY "madness_${_comp}_LIBRARY-NOTFOUND")
  endif()
endforeach()

# done
set(madness_FOUND TRUE)
check_required_components(madness)

